package kdocCxxHTML;

use File::Path;
use File::Basename;

use Carp;
use Ast;
use kdocAstUtil;
use kdocHTMLutil;

use strict;
no strict "subs";

use vars qw/ @clist @docQueue $host $who $now $gentext %toclinks $docBotty
	$lib $rootnode $outputdir $opt $debug *CLASS/;

=head1 kdocCxxHTML

Capabilities required from Ast bit: 

1. Create an inheritance tree

2. Referencing ability: convert a fully qualified class or member name
	to a node reference.

=cut

BEGIN
{
	@clist = ();
	@docQueue = ();

	$host = `uname -n`;	chop $host;
	$who = `whoami`;	chop $who;
	$now = localtime;	chop $now;
	$gentext = "$who\@$host on $now.";

	%toclinks = (
		'Index'		 => 'index.html',
		'Annotated List' => 'index-long.html',
		'Hierarchy'	 => 'hier.html',
		'Globals'	 => 'all-globals.html',
		'Files'		 => 'header-list.html'
	);

	$docBotty =<<EOF
<HR>
	<table>
	<tr><td><small>Generated by: $gentext</small></td></tr>
	</table>
</BODY>
</HTML>
EOF

}

sub writeDoc
{
	( $lib, $rootnode, $outputdir, $opt ) = @_;

	$debug = $main::debug;

	print "Generating HTML documentation. \n" unless $main::quiet;

	mkpath( $outputdir ) unless -f $outputdir;

	makeReferences( $rootnode );
	makeClassList( $rootnode );

	writeGlobalDoc( $rootnode );
	writeClassList( $rootnode );
	writeAnnotatedList( $rootnode );
	writeHier( $rootnode );
	writeHeaderList();

	foreach my $node ( @{$rootnode->{Kids}} ) {
		next unless kdocAstUtil::localComp( $node );

		push @docQueue, $node;
	}

	while( $#docQueue >= 0 ) {
		print "Items in docQueue: $#docQueue \n" if $debug;

		my $node = pop @docQueue;

		writeClassDoc( $node );
	}
}

=head2 writeClassList

	Parameters: rootnode

	Writes out a concise list of classes to index.html

=cut

sub writeClassList
{
	my ( $root ) = @_;

	open(CLIST, ">$outputdir/index.html") 
		|| die "Couldn't create $outputdir/index.html\n";

	newPgHeader( *CLIST{IO}, "$lib Class Index", "", "", \%toclinks );

	if ( defined $root->{DocNode} ) {
		printDoc( $root->{DocNode}, *CLIST, $root, 1 );
	}

	if ( $#clist < 0 ) {
		print "<h2>No classes</h2>";
		# TODO: Perhaps display C-specific index.
	}
	else {
		writeTable( *CLIST{IO}, \@clist, 
			exists $opt->{"html-cols"} ? $opt->{"html-cols"} : 3 );
	}

	print CLIST $docBotty;
	close CLIST;
}



=head2 writeAnnotatedList

	Parameters: rootnode

	Writes out a list of classes with short descriptions to
	index-long.html.

=cut

sub writeAnnotatedList
{
	my ( $root ) = @_;
	my $short;

	open(CLIST, ">$outputdir/index-long.html") 
	|| die "Couldn't create $outputdir/index-long.html\n";

	newPgHeader( *CLIST{IO}, "$lib Annotated List", "", "", \%toclinks );

	print CLIST '<TABLE WIDTH="100%" BORDER=\"0\">';
	
	my $colnum = 0;
	my $colour;
	my $col = 0;

	foreach my $node ( @clist ) {
		print  "undef in clist\n" if !defined $node;

		my $docnode = $node->{DocNode};
		$short = "";

		if( defined $docnode && exists $docnode->{ClassShort} ) {
			$short = deref($docnode->{ClassShort}, $rootnode );
			if( !defined $short ) {
				print $root->{astNodeName}, "has undef short\n";
				next;
			}
				
		}
		
		$col = $col ? 0 : 1;
		$colour = $col ? "" : 'bgcolor="#eeeeee"';

		print CLIST "<TR $colour><TD>", 
			refNameFull( $node, $rootnode ), 
			"</TD><TD>", $short, "</TD></TR>";
	}

	print CLIST "</TABLE>", $docBotty;
	close CLIST;
}


=head2 writeAllMembers

	Parameters: node

	Writes a list of all methods to "full-list-<class file>"

=cut

sub writeAllMembers
{
	my( $node ) = @_;
	my $file = "$outputdir/full-list-".$node->{Ref};
	my %allmem = ();

	kdocAstUtil::allMembers( \%allmem, $node );

	open( ALLMEM, ">$file" ) || die "Couldn't create $file\n";

#	print ALLMEM pageHeader( \%toclinks, esc($node->{astNodeName})
#		." - All Methods" ), "<UL>";

	newPgHeader( *ALLMEM{IO}, 
			$node->{NodeType}." ".esc($node->{astNodeName}).
			": All methods", "", "", \%toclinks );

	my $mem;
	my $col = 0;
	my $colour = "";

	my @memlist = sort { $a->{astNodeName} cmp $b->{astNodeName} } 
			values %allmem;
	writeTable( *ALLMEM{IO}, \@memlist, 3 );

	print ALLMEM "$docBotty";

	close ALLMEM;
}

=head2 writeHier

	Parameters: rootnode

	Writes out the class hierarchy index to hier.html.

=cut

sub writeHier
{
	my ( $root ) = @_;

	open( HIER, ">$outputdir/hier.html") 
		|| die "Couldn't create $outputdir/hier.html\n";

	newPgHeader( *HIER{IO}, "$lib Class Hierarchy", "", "", \%toclinks );


	printNodeHier( $root );

	print HIER $docBotty;
	close HIER;
}

=head3 printNodeHier

	Parameters: node

	Lists all classes that inherit from this node in an unordered list.

=cut

sub printNodeHier 
{
	my( $node ) = @_;
	my $kid;
	my $src = "";

	# non-derived external classes are not printed.
	if ( defined $node->{ExtSource} ) {
		return if !defined $node->{InBy} 
			|| !kdocAstUtil::hasLocalInheritor( $node );

		$src = "<small>(".$node->{ExtSource}.")</small>";
	}

	print HIER "<LI>", refNameFull( $node, $rootnode )," $src</LI>\n" 
			unless $node == $rootnode;

	return if !defined $node->{InBy} || defined $node->{hlisted};

	print HIER "<UL>\n";

	foreach $kid ( sort { $a->{astNodeName} cmp $b->{astNodeName} } 
			@{ $node->{InBy} } ) {
		next if !$main::doPrivate && $kid->{Access} eq "private";
		printNodeHier( $kid );
		$kid->AddProp( 'hlisted', 1 );
	}
	print HIER "</UL>\n";
}

=head2 writeHeaderList

	Generates the header-list.html file, which contains links
	to each processed header. The ARGV list is used.

=cut

sub writeHeaderList
{
	open(HDRIDX, ">$outputdir/header-list.html") 
	|| die "Couldn't create $outputdir/header-list.html\n";

	newPgHeader( *HDRIDX{IO}, "$lib File Index", "", "", \%toclinks );

	foreach my $header ( sort @main::ARGV ) {
		$_ = $header;
		my $htmlheader = $main::striphpath ? basename ( $_ ) : $_;
		my $htmlname = HeaderPathToHTML( $htmlheader );

# convert dashes to double dash, convert path to dash

		print HDRIDX  "\t<LI>",hyper($htmlname,$htmlheader),"</LI>\n";

		writeSrcHTML( "$outputdir/$htmlname", $header );
	}
	
	print HDRIDX "</UL>\n",$docBotty;
	close HDRIDX;
}

=head2 writeClassDoc

	Write documentation for one compound node.

=cut

sub writeClassDoc
{
	my( $node ) = @_;

	print "Enter: $node->{astNodeName}\n" if $debug;
	if( exists $node->{ExtSource} ) {
		warn "Trying to write doc for ".$node->{AstNodeName}.
			" from ".$node->{ExtSource}."\n";
		return;
	}

	my $file = "$outputdir/".$node->{Ref};
	my $docnode = $node->{DocNode};
	my @list = ();
	my $version = undef;
	my $author = undef;

	open( CLASS, ">$file" ) || die "Couldn't create $file\n";

	# Header
	 
	my $source = kdocAstUtil::nodeSource( $node );
	my $short = "";
	my $extra = "";

	if( kdocAstUtil::hasDoc( $node ) ) {
		if ( exists $docnode->{ClassShort} ) {
			$short .= deref($docnode->{ClassShort}, 
				$rootnode).
				" <small>".
				hyper( "#longdesc", "More..." )."</small>";
		}

		if ( exists $docnode->{Deprecated} ) {
			$extra .= '<TR><TH colspan="2">'.
			'Deprecated! use with care</TH></TR>';
		}

		if ( exists $docnode->{Internal} ) {
			$extra .= '<TR><TH colspan="2">'.
			'Internal Use Only</TH></TR>';

		}
		$version = esc($docnode->{Version}) 
			if exists $docnode->{Version};
		$author = esc($docnode->{Author}) 
			if exists $docnode->{Author};
	}

	# pure virtual check
	if ( exists $node->{Pure} ) {
		$extra .= '<TR><TH>Contains pure virtuals</TH></TR>';
	}
	
	# full name, if not in global scope
	if ( $node->{Parent} != $rootnode ) {
		$extra .= tabRow( "Full name", 
			"<code>".refNameEvery( $node, $rootnode )."</code>" );
	}

	# include
	$extra .= tabRow( 'Definition', '<code>#include &lt;'.
		hyper( HeaderPathToHTML( $source ), $source )
		.'&gt;</code>' );

	# template form
	if ( exists $node->{Tmpl} ) {
		$extra .= tabRow( "Template form",
			esc($node->{astNodeName})
			."&lt;".textRef($node->{Tmpl}, $rootnode )."&gt; "
			."</code>" );
	}

	# ancestors
	if ( $node != $rootnode && exists $node->{InList} ) {
		my $comma = "";
		my $out = "";

		foreach my $in ( @{ $node->{InList} } ) {
			my $nref = $in->{Node};

			if( !defined $nref ) {
				# not found
				$out .= $comma.esc($in->{astNodeName});
			}
			elsif ( $nref == $rootnode ) {
				# global
				next;
			}
			else {
				# resolved
				$out .= $comma.refNameFull($nref,$rootnode);
			}
				
			# template type?
			if ( exists $in->{TmplType} ) {
				$out .= " &lt;".wordRef($in->{TmplType},
						$rootnode )."&gt;";
			}


			if ( exists $nref->{ExtSource} ) {
				$out .=" <small>(".$nref->{ExtSource}
				.")</small>";
			}

			$out .= " <small>[".$in->{Type}.  "]</small>" 
				unless $in->{Type} eq "public";

			$comma = ", ";
		}

		$extra .= tabRow( "Inherits", $out ) unless $out eq "";
	}

	# descendants
	if ( $node != $rootnode && exists $node->{InBy} ) {
		my $comma = "";
		my $out = "";

		@list = ();
		kdocAstUtil::inheritedBy( \@list, $node );

		foreach my $in ( @list ) {
			$out .= $comma.refName( $in );

			# external source?
			if ( exists $in->{ExtSource} ) {
				$short .= " <small>(".
				$in->{ExtSource}.")</small>";
			}
	
			$comma = ", ";
		}
		$extra .= tabRow( "Inherited by", $out );
	}

	$extra .= '<TR><TH>'.
		hyper( encodeURL("full-list-".$node->{Ref}),
			"List of all Methods" )."</TH></TR>";


	#### print it

	newPgHeader( *CLASS{IO}, 
		$node->{NodeType}." ".esc($node->{astNodeName}), 
		$short, $extra, \%toclinks );


	if( $#{$node->{Kids}} < 0 ) {
		print CLASS "<center><H4>No members</H4></center>\n";
	}
	else {
		listMethods( $node, "Public Members", "public" );
		listMethods( $node, "Public Slots", "public_slots" );
		listMethods( $node, "Protected Members", "protected" );
		listMethods( $node, "Protected Slots", "protected_slots" );
		listMethods( $node, "Signals", "signals" );

		if ( $main::doPrivate ) {
			listMethods( $node, "Private Members", "private" );
			listMethods( $node, "Private Slots", "private_slots" );
		}
	}

	# long description
	if ( kdocAstUtil::hasDoc( $node ) ) {
		print CLASS "<HR><A NAME=\"longdesc\">",
			"<H2>Detailed Description</H2>";
		printDoc( $docnode, *CLASS, $node, 1 );
	}

	# member doc
	my $kid;
	my ($numref, $ref);

	foreach $kid ( @{$node->{Kids}} ) {
			next if defined $kid->{ExtSource} 
			|| $node->{NodeType} eq "Forward"
			|| (!$main::doPrivate && 
				$kid->{Access} =~ /private/);

		# queue for documenting if compound
		if ( kdocAstUtil::localComp( $kid ) ) {
			print "PushingI: $node->{astNodeName}\n" if $debug;
			push @docQueue, $kid;
		}

		next if !defined $kid->{DocNode};

		if( !exists $kid->{NumRef} ) {
			warn $kid->{astNodeName}, " type ",
			$kid->{NodeType}, " doesn't have a numref\n";
		}

		( $numref = $kid->{NumRef} ) =~ s/^.*?#//g;
		( $ref = $kid->{Ref} ) =~ s/^.*?#//g;

		printMemberName( $kid, $ref, $numref );
		printDoc( $kid->{DocNode}, *CLASS, $node );

		if ( $kid->{NodeType} eq "method" ) {
			$ref = kdocAstUtil::findOverride( $rootnode, $node, 
				$kid->{astNodeName} );
			if ( defined $ref ) {
				print CLASS "<p>Reimplemented from ",
					refName( $ref ), "</p>\n";
			}
		}
	}

	# done

	if ( defined $version || defined $author ) {
		print CLASS "<HR><UL>",
			defined $version ? 
				"<LI><i>Version</i>: $version</LI>" : "",
			defined $author ? 
				"<LI><i>Author</i>: $author</LI>" : "",
			"<LI><i>Generated</i>: $gentext</LI></UL>",
			"</BODY></HTML>\n";
	}
	else {
		print CLASS $docBotty;
	}

	close CLASS;

	# full member list

	writeAllMembers( $node );
}

sub writeGlobalDoc
{
	my( $node ) = @_;
	my $file = "$outputdir/all-globals.html";
	my $docnode = $node->{DocNode};
	my $hasdoc = exists $node->{DocNode} ? 1 : 0;
	my @list = ();
	my $cumu = Ast::New( "nodelist" );
	my $kid;

	# make a list of nodes by file
	foreach $kid ( @{$node->{Kids}} ) {
		next if exists $kid->{ExtSource} 
			|| exists $kid->{Compound}
			|| (!$main::doPrivate && 
				$kid->{Access} =~ /private/);

		$cumu->AddPropList( kdocAstUtil::nodeSource( $kid ), $kid )
			unless !exists $kid->{Source};
	}

	open( CLASS, ">$file" ) || die "Couldn't create $file\n";

	newPgHeader( *CLASS{IO}, $lib." Globals", "", "", \%toclinks );


	@list = sort keys %$cumu;

	foreach $file ( @list ) {
		next if $file eq "astNodeName";

		listMethods( $node, $file, "", $cumu->{$file} );
	}

	# member doc
	my ($numref, $ref);

	foreach $file ( @list ) {
		next if $file eq "astNodeName";
		
		foreach $kid ( @{$cumu->{$file}} ) {
			next if exists $kid->{ExtSource} 
			|| exists $kid->{Compound}
			|| !exists $kid->{DocNode}
			|| (!$main::doPrivate && 
					$kid->{Access} =~ /private/);

			if( !exists $kid->{NumRef} ) {
				warn $kid->{astNodeName}, " type ",
				$kid->{NodeType}, " doesn't have a numref\n";
			}

			( $numref = $kid->{NumRef} ) =~ s/^.*?#//g;
				( $ref = $kid->{Ref} ) =~ s/^.*?#//g;

				printMemberName( $kid, $ref, $numref );

			print CLASS "<p><small><code>#include &lt;",
				kdocAstUtil::nodeSource( $kid ), 
				"&gt;</code></small></p>";

			printDoc( $kid->{DocNode}, *CLASS, $node );
		}
	}

	print CLASS $docBotty;
	close CLASS;
}

sub listMethods
{
	my( $class, $desc, $vis, $nodes ) = @_;
	my $name;
	my $type;
	my $flags;
	my @n=();
	
	if ( !defined $nodes ) { 
		kdocAstUtil::findNodes( \@n, $class->{Kids}, 
			"Access", $vis );
		$nodes = \@n;
	}

	return if ( $#{$nodes} < 0 );

print CLASS<<EOF;
<H2>$desc</H2>
<UL>
EOF
	foreach my $m ( @$nodes ) {
		next if exists $m->{ExtSource};
		if( exists $m->{Compound} ) {
			# compound docs not printed for rootnode
			next if $class eq $rootnode; 

			$name = refName( $m );
		}
		elsif( exists $m->{DocNode} ) {
			# compound nodes have their own page
			$name = refName( $m,  'NumRef' );
		} else {
			$name = esc( $m->{astNodeName} );
		}

		$type = $m->{NodeType};

		print CLASS "<LI>";

		if( $type eq "var" ) {
			print CLASS esc( $m->{Type}), 
				" <b>", $name,"</b>\n";
		}
		elsif( $type eq "method" ) {
			$flags = $m->{Flags};

			if ( !defined $flags ) {
				warn "Method ".$m->{astNodeName}.
					" has no flags\n";
			}

			$name = "<i>$name</i>" if $flags =~ /p/;
			my $extra = "";
			$extra .= "virtual " if $flags =~ "v";
			$extra .= "static " if $flags =~ "s";

			print CLASS $extra, esc($m->{ReturnType}),
				" <b>", $name, "</b> (", 
				esc($m->{Params}), ") ", 
				$flags =~ /c/ ? " const\n": "\n";
		}
		elsif( $type eq "enum" ) {
			print CLASS "enum <b>", $name, "</b> {",
				esc($m->{Params}),"}\n";
		}
		elsif( $type eq "typedef" ) {
			print CLASS "typedef ", 
				esc($m->{Type}), " <b>",
				$name,"</b>";
		}
		else { 
			# unknown type
			print CLASS esc($type), " <b>",
				$name,"</b>\n";
		}

		print CLASS "</LI>\n";
	}

print CLASS<<EOF;
</UL>
EOF

}

=head2 printIndexEntry

	Parameters: member node

	Prints an index entry for a single node.

	TODO: stub

=cut

sub printIndexEntry
{
	my ( @node ) = @_;
}

=head2 printMemberName

	Parameters: member node, names...

	Prints the name of one member, customized to type. If names are
	specified, a name anchor is written for each one.

=cut

sub printMemberName
{
	my $m = shift;

	my $name = esc( $m->{astNodeName} );
	my $type = $m->{NodeType};
	my $ref;
	my $flags = undef;

	foreach $ref ( @_ ) {
		print CLASS "<A NAME=\"", $ref, "\"></A>";
	}

	print CLASS '<table width="100%"><tr bgcolor="#eeeeee"><td><strong>';

	if( $type eq "var" ) {
		print CLASS textRef($m->{Type}, $rootnode ), 
		" <b>", $name,"</b>\n";
	}
	elsif( $type eq "method" ) {
		$flags = $m->{Flags};
		$name = "<i>$name</i>" if $flags =~ /p/;

		print CLASS textRef($m->{ReturnType}, $rootnode ),
		" <b>", $name, "</b> (", 
		textRef($m->{Params}, $rootnode ), ")\n";
	}
	elsif( $type eq "enum" ) {
		print CLASS "enum <b>", $name, "</b> {",
		esc($m->{Params}),"}\n";
	}
	elsif( $type eq "typedef" ) {
		print CLASS "typedef ", 
		textRef($m->{Type}, $rootnode ), " <b>", $name,"</b>";
	}
	else {
		print CLASS $name, " <small>(", esc($type), ")</small>";
	}

	print CLASS "</strong></td></tr></table>";

# extra attributes
	my @extra = ();

	if( !exists $m->{Access} ) {
		print "Member without access:\n";
		kdocAstUtil::dumpAst( $m );
	}

	($ref = $m->{Access}) =~ s/_slots//g;

	push @extra, $ref
		unless $ref =~ /public/
		|| $ref =~ /signal/;

	if ( defined $flags ) {
		my $f;
		my $n;
		foreach $f ( split( "", $flags ) ) {
			$n = $main::flagnames{ $f };
			warn "flag $f has no long name.\n" if !defined $n;
			push @extra, $n;
		}
	}

	if ( $#extra >= 0 ) {
		print CLASS " <small>[", join( " ", @extra ), "]</small>";
	}

	print CLASS "</p>";
	
# finis
}



=head2 makeClassList

	Parameters: node

	fills global @clist with a list of all direct, non-external
	compound children of node.

=cut

sub makeClassList
{
	my ( $rootnode ) = @_;

	@clist = ();

	foreach my $node ( @ {$rootnode->{Kids}} ) {
		if ( !defined $node ) {
			print "makeClassList: undefined child in rootnode!\n";
			next;
		}

		push( @clist, $node ) if kdocAstUtil::localComp( $node );
	}

	@clist = sort { $a->{astNodeName} cmp $b->{astNodeName}  }
			@clist;
}

sub writeSrcHTML
{
	my ( $outfile, $infile ) = @_;

	open ( OUT, ">$outfile" ) || die "Couldn't open $outfile for".
			"writing.\n";

	newPgHeader( *OUT{IO}, "Source: $infile", "", "", \%toclinks );
	makeHeader( *OUT{IO}, $infile );

	print OUT $docBotty;
	close OUT;
}

1;
