package kdoctexi;

# kdoctexi.pm -- C++ TexInfo output module for KDOC.
#		Copyright (C) 1998, Bernd Gehrmann
# $Id: kdoctexi.pm,v 1.8 1999/11/02 09:13:16 ssk Exp $

# The KDOC package is distributed under the GNU Public License.

use Ast;
use kdocAstUtil;
use File::Path;
use File::Basename;

use strict;

use vars qw/ $lib $rootnode $outputdir @clist $depth /;

=head1 kdoctexi
	TexInfo output module.
=cut
    
sub writeDoc 
{
        ( $lib, $rootnode, $outputdir ) = @_;
  
        print "Generating texinfo documentation.\n" unless $main::quiet;

	mkpath( $outputdir ) unless -f $outputdir;

        makeClassList( $rootnode );
        writeMain();
        writeHierarchy();
        writeOverview();
        writeClasses();

}



sub makeClassList
{
	my ( $rootnode ) = @_;

	@clist = ();

	foreach my $node ( @ {$rootnode->{Kids}} ) {
		if ( !defined $node ) {
			print "makeClassList: undefined child in rootnode!\n";
			next;
		}

		push( @clist, $node ) unless exists $node->{ExtSource}
				|| !exists $node->{Compound};
	}

	@clist = sort { $a->{astNodeName} cmp $b->{astNodeName}  }
			@clist;
}



sub writeMain {

        open( TEXMAIN, ">$outputdir/$lib.texi" ) 
	  || die "Couldn't write to $outputdir/$lib-main.texi\n";

        print TEXMAIN<<EOF;
\\input texinfo
\@c %**start of header
\@comment ----- automatically generated - do not edit! -----
\@afourpaper
\@setfilename $lib.info
\@settitle $lib documentation
\@headings on
\@setchapternewpage on
\@c %**end of header

\@include $lib-hier.tex
\@include $lib-overvw.tex
\@include $lib-inc.tex

\@page
\@comment \@printindex fn
\@shortcontents
\@bye
EOF

        close( TEXMAIN );
}



sub writeHierarchy 
{
        open( TEX, ">$outputdir/$lib-hier.tex" ) 
	  || die "Couldn't write to $outputdir/$lib-hier.tex\n";
	
	print TEX "\@comment ----- automatically generated - do not edit! -----\n";
	print TEX "\@unnumbered $lib Class Hierarchy\n\n";
	$depth = 0;
	writeNode( $rootnode );
	
	close TEX;
}



sub writeNode 
{
        my ( $node ) = @_;
	my ( @children );
	
	if ( $node ne $rootnode ) {

                my $name = $node->{astNodeName};
		my $source = $node->{ExtSource};

		print TEX "\@code{$name}";
		if ( defined $source ) {
			print TEX " (from $source)";
		}
		print TEX "\n";
	}
	
	my $bulletType = undef;
	if ( defined $node->{InBy} ) {
		$depth++;
		if ($depth == 1) {
			$bulletType = "\@bullet";
		}
		elsif ($depth == 2) {
			$bulletType = "*";
		}
		elsif ($depth == 3) {
			$bulletType = "+";
		}
		elsif ($depth > 3) {
			$bulletType = "-";
		}
		# Insert your favorite bullet for deeper hierarchies :-)
		print TEX "\@itemize $bulletType\n";
		
		foreach my $child ( 
				sort { $a->{astNodeName} cmp $b->{astNodeName} }
                                @{ $node->{InBy} } ) {
			print TEX "\@item\n";
			writeNode( $child );
		}
		
		print TEX "\@end itemize\n";
		$depth--;
	}
}



sub writeOverview 
{
	open( TEX, ">$outputdir/$lib-overvw.tex" ) 
	  || die "Couldn't write to $outputdir/$lib-classes.tex\n";
	
	print TEX "\@comment ----- automatically generated - do not edit! -----\n";
	print TEX "\@unnumbered $lib Library Overview\n\n";
	print TEX "\@table \@asis\n";
	
	foreach my $node ( @clist ) {

                my $docnode = $node->{DocNode};

		my $short = "";
		if ( defined $docnode && exists $docnode->{ClassShort} ) {
		        $short = formatText($docnode->{ClassShort});
		}

		my $className = $node->{astNodeName};

                # The following is really a dirty hack
                # Find something better!
		print TEX "\@item \@code{$className ";
	        print TEX "@ @ @ @ @ @ @ @ @ @ @ @ @ }", "\n";
		print TEX escape($short), "\n";
	}
						  
	print TEX "\@end table\n";
	close TEX;
}



sub writeClasses 
{
	open( TEXINC, ">$outputdir/$lib-inc.tex" ) 
	  || die "Couldn't write to $outputdir/$lib-inc.tex\n";
	print TEXINC "\@comment ----- automatically generated - do not edit! -----\n";
	
	foreach my $node ( @clist ) {

		next if !defined $node->{Compound} 
		  || defined $node->{ExtSource};

		my $className = $node->{astNodeName};

		print TEXINC "\@include $className.tex\n";
		open( TEX, ">$outputdir/$className.tex" ) 
		  || die "Couldn't write to $outputdir/$className.tex\n";
		print TEX "\@comment ----- automatically generated - do not edit! -----\n";

		writeClass( $node );
		close TEX;
	}
	
	close TEXINC;
}



sub writeClass 
{
	my ( $node ) = @_;

	my $docnode = $node->{DocNode};
	my $header = kdocAstUtil::nodeSource($node);
	my $className = $node->{astNodeName};
	
	my $classType = (exists $node->{Tmpl} )? "Template" : "Class";
	
	print TEX<<EOF;
\@unnumbered \@code\{$className\} $classType Reference \@i\{($header)\}
\@findex $className
EOF
	
	if ( exists $docnode->{Deprecated} ) {
		print TEX "\@noindent\n\@b{Deprecated Class}\n";
	}
	
	if ( exists $docnode->{Internal} ) {
		print TEX "\@noindent\n\@b{For internal use only}\n";
	}

	if ( exists $node->{Pure} ) {
		print TEX "\@noindent\n\@b{Contains pure virtuals}\n";
	}
	
	if ( $classType eq "template" ) {
		print TEX "\@table \@asis\n\@item Template form:\n",
		"\@code{template <", escape( $node->{Tmpl} ),
		"> ", $className, "}\n\@end table\n";
	}
	
	if ( $node != $rootnode && exists $node->{InList} ) {
		my $ref = $node->{InList};
		my $out = "";
		
		my $comma = "\@table \@asis\n\@item Inherits:\n";
		
		foreach my $ancestor ( @{ $node->{InList} } ) {
			my $n = $ancestor->{Node};
			next if $n == $rootnode;

			my $ancName;
			my $ancSource = "";

			if ( defined $n ) {
				$ancName = $n->{astNodeName};
	                        $ancSource = "(".$n->{ExtSource}.")";
			}
			else {
				$ancName = " (".$ancestor->{astNodeName}.")";
			}

			$out .= $comma . escape( $ancName ). $ancSource;
			$comma = ", ";
		
		}
		
		print TEX $out."\n\@end table\n" unless $out eq "";
	}
	
	if ( $node != $rootnode && exists $node->{InBy} ) {
		
		print TEX "\@table \@asis\n\@item Inherited by:\n";
		
		my @list = ();
		kdocAstUtil::inheritedBy( \@list, $node );

		my ( $pcount ) = 0;
		foreach my $inheritor ( @list ) {
			next if $inheritor == $rootnode;
			
			my $inhName = $inheritor->{astNodeName};
			my $inhSource = $inheritor->{ExtSource};

			print TEX ", " unless $pcount == 0;
			$pcount = 1;
			
			print TEX $inhName;
			if ( defined $inhSource ) {
				print TEX " ($inhSource)";
			}
			
		}
		
		print TEX "\n\@end table\n";
	}

        my $author   = $node->{Author};
        my $version  = $node->{Version};
        my $classSee = $node->{See};

	if( $author ne "" || $version ne "" || $classSee ne "" ) {
		my $escVersion = escape( $version );
		my $escAuthor = escape( $author );
		$escAuthor =~ s/<(.+\@\@.+)>/\@email{$1}/mg;
		$escAuthor =~ s/\((.+\@\@.+)\)/\@email{$1}/mg;
		my $escSee = escape( $classSee );
		print TEX "\@table \@asis\n";
		print TEX "\@item Version:\n$escVersion\n" unless $escVersion eq "";
		print TEX "\@item Author:\n$escAuthor\n" unless $escAuthor eq "";		
		print TEX "\@item See Also:\n$escSee\n" unless $escSee eq "";		
		print TEX "\@end table\n";
	}
	
	writeDescription( $docnode );
	
	writeMembers( $node, "Public Members", "public" );
	writeMembers( $node, "Public Slots", "public_slots" );
	writeMembers( $node, "Protected Members", "protected" );
	writeMembers( $node, "Protected Slots", "protected_slots" );
	writeMembers( $node, "Signals", "signals" );
	
	if( $main::doPrivate ) {
		writeMembers( $node, "Private Members", "private" );
		writeMembers( $node, "Private Slots", "private_slots" );
	}
	
}



sub writeMembers 
{
	my( $classnode, $desc, $vis ) = @_;
	
	my @nodes = ();
	kdocAstUtil::findNodes( \@nodes, $classnode->{Kids}, "Access", $vis );
	if( $#nodes < 0 ) { return; }
	
	print TEX "\@unnumberedsec $desc\n";
	
	foreach my $member ( @nodes ) {
		writeMember( $classnode, $member );
	}
	
}



sub writeMember 
{
	my ( $classnode, $member ) = @_;
	my $className = $classnode->{astNodeName};
	my $memberName = $member->{astNodeName};
        my $escName = escape( $memberName );
	my $type = $member->{NodeType};
        $_ = $type;
        SWITCH: {
		
		if( /^method/ ) {
			my $escRetType = escape( $member->{ReturnType} );
			my $escParams = escape( $member->{Params} );
			my $flags = $member->{Flags};

			if ( $memberName eq $className ) {
				print TEX "\@deftypefn Constructor {} $escName {($escParams)}\n";
			}
			elsif ( $memberName eq "~$className" ) {
				print TEX "\@deftypefn Destructor {} $escName {($escParams)}\n";
			}
                        elsif ( $flags =~ /s/ ) {
				print TEX "\@deftypefn {Static Method} {$escRetType} $escName ",
				"{($escParams)}\n";
			}
			else {
				print TEX "\@deftypefn Method {$escRetType} $escName ",
				"{($escParams)}", $flags =~ /c/? " const\n" : "\n";
			}
			
			my $ref = kdocAstUtil::findOverride( $rootnode, 
					$classnode, $member->{astNodeName} );
			if ( defined $ref ) {
			        print CLASS "Reimplemented from $ref\n";
			}

			writeMemberInfo( $member->{DocNode} );
			
                        my @paramlist = ();
                        kdocAstUtil::findNodes( \@paramlist, $member->{Text},
                                                "NodeType", "Param" );

			if ( $#paramlist >= 0 ) {
			        print TEX "\@noindent\nParameters:\n",
			          "\@multitable \@columnfractions .1 .15 .75\n";
	                        foreach my $paramnode ( @paramlist ) {
					print TEX "\@item \@tab ",
					escape($paramnode->{Name}), " \@tab ",
					escape($paramnode->{astNodeName}), "\n";
					
				}
					print TEX "\@end multitable\n";
			}
				
			my $returns = $member->{Returns};
		        print TEX "\@noindent\nReturns: ", 
				escape($returns), "\n" unless $returns eq "";
				
                        my $exceptions = $member->{Throws};
			if ( $exceptions ne "" ) {
			        print TEX "\@noindent\n\@table ",
					"\@asis\n\@item Throws:\n",
					escape($exceptions), "\n\@end table\n";
			}
				
			print TEX "\@end deftypefn\n\n";
			last SWITCH;
		} # /^method/
		
		if ( /^var/ ) {
			my $type = $member->{Type};
			print TEX "\@deftypevar", escape($type), " $escName\n";
			writeMemberInfo( $member->{DocNode} );
			print TEX "\@end deftypevar\n";
			last SWITCH;
		} # /^var/
		
		if( /^typedef/ ) {
			my $escType = escape( $member->{Type} );
			print TEX "\@deftp {typedef} {$escType} $escName\n";
			writeMemberInfo( $member->{DocNode} );
			print TEX "\@end deftp\n";
			last SWITCH;
		} # /^typedef/
		
		if ( /^enum/ ) {
			# I'm not really convinced that it makes sense to list all constants.
			# Normally, I would expect that all constants are explained in the
			# doc comment, embedded in a <ul>..</ul> list.
			my $escConstants = escape( $member->{Params} );
			$escConstants =~ s/\n+/\n/g;
                        my $enumName = ($escName =~ /^\s*$/)? 
				"(Anonymous)" : $escName;
			print TEX "\@deftp Enumeration \@code{enum} $enumName\n";
			print TEX "\@example\nenum $enumName \@{\n",
			"$escConstants\@};\n\@end example\n";
			writeMemberInfo( $member->{DocNode} );
			print TEX "\@end deftp\n";
			last SWITCH;
		} # /^enum/
		
	} # SWITCH
	
}



sub writeDescription 
{
        my ( $docnode ) = @_;

        my $text = $docnode->{Text};
	if ( $text eq "") { return; }

        my $lasttype = "";	
	my $type = undef;
	foreach my $node ( @$text ) {
                $type = $node->{NodeType};
                my $name = $node->{astNodeName};

		if ( $lasttype eq "ListItem" && $type ne $lasttype ) {
                        print TEX "\@end itemize\n";
		}

		if ( $type eq "Pre" ) {
                        print TEX "\@example\n";
		        print TEX escape($name);
                        print TEX "\n\@end example\n";
		}
                elsif ( $type eq "DocText" || $type eq "Ref" ) {
                        print TEX formatText($name);
		}
                elsif ( $type eq "ParaBreak" ) {
                        print TEX "\n\n";
	        }
		elsif ( $type eq "ListItem" ) {
		        if ( $lasttype ne "ListItem" ) {        
                                print TEX "\@itemize \@bullet\n";
			}
                        print TEX "\@item ", formatText($name), "\n";
		}

		$lasttype = $type;

	}

        if( $type eq "ListItem" ) {
	        print TEX "\@end itemize\n"; 
        }

	print TEX "\n";
	
}



sub writeMemberInfo 
{

        my ( $docnode ) = @_;

	writeDescription( $docnode );
	
#	if( $MethDeprecated ) {
#		print TEX "\@noindent\n\@b{Deprecated Member}\n";
#	}
#	
#	if( $MethInternal ) {
#		print TEX "\@noindent\n\@b{For internal use only}\n";
#	}
	
}


# This is for conversion of general, non-preformatted text

sub formatText 
{
	my ( $text ) = @_;
	
	$text = escape($text);
	# After escape, @ is @@
	$text =~ s/\@\@ref\s+([\w\d_~]+)(#|::)([\w\d_~]+)/$1::$3/g;
        $text =~ s/\@\@ref\s+(#|::)?([\w\d_~]+)/$2/g;
	$text =~ s/\@\@see\s+([\w\d_~]+)(#|::)([\w\d_~]+)/$1::$3/g;
        $text =~ s/\@\@see\s+(#|::)?([\w\d_~]+)/$2/g;
	$text =~ s/<em>/\@emph\{/ig;
	$text =~ s/<\/em>/\}/ig;
	$text =~ s/<strong>/\@strong\{/ig;
        $text =~ s/<\/strong>/\}/ig;
        $text =~ s/<dfn>/\@dfn\{/ig;
        $text =~ s/<\/dfn>/\}/ig;
        $text =~ s/<code>/\@code\{/ig;
        $text =~ s/<\/code>/\}/ig;
        $text =~ s/<samp>/\@samp\{/ig;
        $text =~ s/<\/samp>/\}/ig;
        $text =~ s/<kbd>/\@kbd\{/ig;
        $text =~ s/<\/kbd>/\}/ig;
        $text =~ s/<var>/\@var\{/ig;
        $text =~ s/<\/var>/\}/ig;
        $text =~ s/<cite>/\@cite\{/ig;
        $text =~ s/<\/cite>/\}/ig;
        $text =~ s/<i>/\@i\{/ig;
        $text =~ s/<\/i>/\}/ig;
        $text =~ s/<b>/\@b\{/ig;
        $text =~ s/<\/b>/\}/ig;
        $text =~ s/<tt>/\@t\{/ig;
        $text =~ s/<\/tt>/\}/ig;

        $text =~ s/<br>/\@*/ig;
        $text =~ s/<p>/\n/ig;
        $text =~ s/<\/p>//ig;
			 
        $text =~ s/&lt;/</g;
        $text =~ s/&gt;/>/g;
        $text =~ s/&amp;/&/g;
        $text =~ s/&nbsp;/\@ /g;
        $text =~ s/&copy;/\@copyright{}/g;

# Support for other languages?
        $text =~ s/&auml;/\@"a/g;
        $text =~ s/&ouml;/\@"o/g;
        $text =~ s/&uuml;/\@"u/g;
        $text =~ s/&Auml;/\@"A/g;
        $text =~ s/&Ouml;/\@"O/g;
        $text =~ s/&Uuml;/\@"U/g;
        $text =~ s/&szlig;/\@ss/g;

        $text =~ s/^\s*<blockquote>\s*$/\@quotation/mig;
        $text =~ s/^\s*<\/blockquote>\s*$/\@end quotation/mig;
        $text =~ s/^\s*<ul>\s*$/\@itemize \@bullet/mig;
        $text =~ s/^\s*<\/ul>\s*$/\@end itemize/mig;
        $text =~ s/^\s*<ol>\s*$/\@enumerate/mig;
        $text =~ s/^\s*<\/ol>\s*$/\@end enumerate/mig;
        $text =~ s/^\s*<dl>\s*$/\@table \@asis/mig;
        $text =~ s/^\s*<\/dl>\s*$/\@end table/mig;
        $text =~ s/^\s*<li>/\@item /mig;
        $text =~ s/^\s*<\/li>//mig;
        $text =~ s/^\s*<dt>/\@item /mig;
        $text =~ s/^\s*<\/dt>//mig;
        $text =~ s/^\s*<dd>//mig;
        $text =~ s/^\s*<\/dd>//mig;

        return $text;
}


# This is for preformatted or similar text
sub escape 
{
	my( $text ) = @_;
	
	$text =~ s/([\@\{\}])/\@$1/g;

# i18n support
#  $text =~ s//\@"a/g;
#  $text =~ s//\@"o/g;
#  $text =~ s//\@"u/g;
#  $text =~ s//\@"A/g;
#  $text =~ s//\@"O/g;
#  $text =~ s//\@"U/g;
#  $text =+ s//\@ss{}/g;

	return $text;
}

#

sub deref
{
	my ( $str ) = @_;
	my $out = "";
	my $text;

	$str =~ s/\@\@ref\s+([\w\d_~]+)(#|::)([\w\d_~]+)/$1::$3/g;
        $str =~ s/\@\@ref\s+(#|::)?([\w\d_~]+)/$2/g;
        $out = $str;

#	foreach $text ( split (/(\@ref\s+[\w:#]+)/, $str ) ) {
#		if ( $text =~ /\@ref\s+([\w:#]+)/ ) {
#			$out .= $1;
#		}
#		else {
#			$out .= $text;
#		}
#	}
	
	return $out;
}

1;
